/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR average_throughput PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received average_throughput copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Junling Bu <linlinjavaer@gmail.com>
 */
#include "ns3/node.h"
#include "ns3/packet.h"
#include "ns3/simulator.h"
#include "ns3/log.h"
#include "ns3/config.h"
#include "ns3/node-container.h"
#include "ns3/net-device-container.h"
#include "ns3/yans-wifi-helper.h"
#include "ns3/mobility-helper.h"
#include "ns3/rng-seed-manager.h"
#include "ns3/command-line.h"
#include "ns3/mobility-model.h"
#include <iostream>
#include <algorithm>
#include "ns3/string.h"
#include "ns3/double.h"
#include "ns3/object-map.h"
#include "ns3/node-list.h"
#include "ns3/seq-ts-header.h"
#include "ns3/regular-wifi-mac.h"
#include "ns3/constant-velocity-mobility-model.h"
#include "ns3/wave-net-device.h"
#include "ns3/wave-mac-helper.h"
#include "ns3/wave-helper.h"

using namespace ns3;

NS_LOG_COMPONENT_DEFINE ("IEEE16094");

class StatsTag : public Tag
{
public:
  StatsTag (void)
    : m_packetId (0),
      m_sendTime (Seconds (0))
  {
  }
  StatsTag (uint32_t packetId, Time sendTime)
    : m_packetId (packetId),
      m_sendTime (sendTime)
  {
  }
  virtual ~StatsTag (void)
  {
  }

  uint32_t GetPacketId (void)
  {
    return m_packetId;
  }
  Time GetSendTime (void)
  {
    return m_sendTime;
  }

  static TypeId GetTypeId (void);
  virtual TypeId GetInstanceTypeId (void) const;
  virtual uint32_t GetSerializedSize (void) const;
  virtual void Serialize (TagBuffer i) const;
  virtual void Deserialize (TagBuffer i);
  virtual void Print (std::ostream &os) const;

private:
  uint32_t m_packetId;
  Time m_sendTime;
};
TypeId
StatsTag::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::StatsTag")
    .SetParent<Tag> ()
    .AddConstructor<StatsTag> ()
  ;
  return tid;
}
TypeId
StatsTag::GetInstanceTypeId (void) const
{
  return GetTypeId ();
}
uint32_t
StatsTag::GetSerializedSize (void) const
{
  return sizeof (uint32_t) + sizeof (uint64_t);
}
void
StatsTag::Serialize (TagBuffer i) const
{
  i.WriteU32 (m_packetId);
  i.WriteU64 (m_sendTime.GetMicroSeconds ());
}
void
StatsTag::Deserialize (TagBuffer i)
{
  m_packetId = i.ReadU32 ();
  m_sendTime = MicroSeconds (i.ReadU64 ());
}
void
StatsTag::Print (std::ostream &os) const
{
  os << "packet=" << m_packetId << " sendTime=" << m_sendTime;
}

// it is useless here we can use any value we like
// because we only allow to send and receive one type packets
#define Packet_Number 0x8888

// the transmission range of a device should be decided carefully,
// since it will affect the packet delivery ratio and packet reception ratio
// we suggest users use wave-transmission-range.cc to get this value
#define Device_Transmission_Range 250

#define CONFIG_CONTINUOUS               1
#define CONFIG_ALTERNATING				2
#define CONFIG_ALTERNATING_80_20	3
#define CONFIG_ALTERNATING_20_80	4

class Ieee16094Experiment
{
public:
  Ieee16094Experiment (void);

  bool Configure (int argc, char **argv);
  void Usage (void);
  void Run (void);

private:

  class BroadcastPacket
    {
    public:
  	  uint32_t src;
  	  uint32_t app;
  	  uint32_t freq;
  	  // the neighbors when broadcasting this packet;
  	  std::vector<uint32_t> sendNeighbors;
  	  // the neighbors who receive this broadcasted packet successfully
  	  std::vector<uint32_t> recvNeighbors;
    };

  void CreateNodes (void);
  void CreateWaveDevice (void);
  void SetupMobility (void);
  void AssignChannelAccess (void);

  // this is an application with frequency transmit
  void InstallApplication (uint32_t protocol);

  void Send (Ptr<WaveNetDevice> sender, uint32_t size, uint32_t app, uint32_t freq);
  bool Receive (Ptr<NetDevice> dev, Ptr<const Packet> pkt, uint16_t mode, const Address &sender);

  void InitStats (void);
  void Stats (uint32_t randomNumber);
  void StatQueuedPackets (void);
  void StatQueue (Ptr<RegularWifiMac> mac, std::string queue);

  NodeContainer nodes;
  NetDeviceContainer devices;

  uint32_t nodesNum;          // current nodes in 4-lines and 1km
  uint32_t freq;              // the frequency for sending packets
  uint32_t simulationTime;    // simulation time
  uint32_t size;              // the size of packet
  uint32_t cwmin;  // Minimum contention window
  // see CONFIG_CONTINUOUS
  // see CONFIG_ALTERNATING
  // see CONFIG_ALTERNATING_80_20
  // see CONFIG_ALTERNATING_20_80
  uint32_t config;

  // used for identifying received packet
  uint32_t sequence;

  uint32_t random_stream;

  // (packetId, packet)
  std::map<uint32_t,  BroadcastPacket *> broadcastPackets;

  class SendStat
  {
  public:
    uint32_t inCchi;
    uint32_t inCgi;
    uint32_t inSchi;
    uint32_t inSgi;
    SendStat (): inCchi (0), inCgi (0), inSchi (0), inSgi (0){}
  };

  SendStat sends;

  uint32_t queues;

  uint32_t pdr_sends;
   uint32_t pdr_recvs;
   // for PAD
   uint64_t pad_delay;    // uint64_t pad_number = pdr_recvs;
   // for PRR (packet receive ratio)
   uint32_t prr_sends;
   uint32_t prr_recvs;

  uint32_t run;
  double pdr,prr,delay,system_throughput,average_throughput;
};

Ieee16094Experiment::Ieee16094Experiment (void)
  : nodesNum (20),           // 20 vehicles in 1km * 4-line
    freq (10),               // 10Hz, 100ms send one non-safety packet
    simulationTime (1),     // make it run 1s
    size (400),             // packet size
    cwmin (15),
    config (1),
    sequence (0),
    random_stream (1),
    queues (0),
    pdr_sends(0),
    pdr_recvs (0),
    pad_delay (0),
    prr_sends (0),
    prr_recvs (0),
    run (1),
    pdr (0),
    prr (0),
    delay (0),
    system_throughput (0),
    average_throughput (0)
{
}

bool
Ieee16094Experiment::Configure (int argc, char **argv)
{
  CommandLine cmd;
  cmd.AddValue ("nodes", "Number of nodes.", nodesNum);
  cmd.AddValue ("time", "Simulation time, s.", simulationTime);
  cmd.AddValue ("size", "Size of safety packet, bytes.", size);
  cmd.AddValue ("frequency", "Frequency of sending safety packets, Hz.", freq);
  cmd.AddValue ("run", "to make the result more credible, make simulation run many times with different random number).", run);
  cmd.AddValue ("cwmin", "Minimum contention window.", cwmin);
  cmd.AddValue ("config", "Configuration.", config);

  cmd.Parse (argc, argv);
  return true;
}

void
Ieee16094Experiment::Usage (void)
{
  std::cout << "usage:"
		    << "./waf --run=\"ieee16094 --nodes=20 --time=100 --size=200 --frequency=10  \""
            << std::endl;
}

void
Ieee16094Experiment::CreateNodes ()
{
  NS_LOG_FUNCTION (this);
  NS_ASSERT (nodesNum != 0);
  nodes = NodeContainer ();
  nodes.Create (nodesNum);
  NS_LOG_DEBUG ("create " << nodesNum << " nodes");
}

void
Ieee16094Experiment::CreateWaveDevice (void)
{
  NS_LOG_FUNCTION (this);
  // the transmission range is about 250m refer to "wave-transmission-range.cc"
  YansWifiChannelHelper waveChannel;
  waveChannel.SetPropagationDelay ("ns3::ConstantSpeedPropagationDelayModel");
  waveChannel.AddPropagationLoss ("ns3::TwoRayGroundPropagationLossModel",
 			  	  	  	  	  	  "Frequency", DoubleValue (5.89e9),
 			  	  	  	  	      "HeightAboveZ", DoubleValue (0.5));
   YansWavePhyHelper wavePhy =  YansWavePhyHelper::Default ();
   wavePhy.Set("TxPowerStart",  DoubleValue (16));
   wavePhy.Set("TxPowerEnd",  DoubleValue (16));
   wavePhy.SetChannel (waveChannel.Create ());
   QosWaveMacHelper waveMac = QosWaveMacHelper::Default ();
   WaveHelper waveHelper = WaveHelper::Default ();
   devices = waveHelper.Install (wavePhy, waveMac, nodes);

	  for (uint32_t i = 0; i != devices.GetN (); ++i)
	    {
		  Ptr<WaveNetDevice> device = DynamicCast<WaveNetDevice>(devices.Get(i));
	      device->SetReceiveCallback (MakeCallback (&Ieee16094Experiment::Receive,this));

	      std::map<uint32_t, Ptr<OcbWifiMac> > macs = device->GetMacs();

		      for (std::map<uint32_t, Ptr<OcbWifiMac> >::iterator m = macs.begin(); m != macs.end(); ++m)
		      {
	   	      Ptr<RegularWifiMac> rmac = DynamicCast<RegularWifiMac>(m->second);
		       PointerValue ptr;
		 	  rmac->GetAttribute ("DcaTxop", ptr);
		 	  Ptr<DcaTxop> dcaTxop = ptr.Get<DcaTxop> ();
		 	  dcaTxop->SetMinCw(cwmin);

		 	  rmac->GetAttribute ("VO_EdcaTxopN", ptr);
		 	  Ptr<EdcaTxopN> vo_edcaTxopN = ptr.Get<EdcaTxopN> ();
		 	  vo_edcaTxopN->SetMinCw(cwmin);

		 	  rmac->GetAttribute ("VI_EdcaTxopN", ptr);
		 	  Ptr<EdcaTxopN> vi_edcaTxopN = ptr.Get<EdcaTxopN> ();
		 	  vi_edcaTxopN->SetMinCw(cwmin);

		 	  rmac->GetAttribute ("BE_EdcaTxopN", ptr);
		 	  Ptr<EdcaTxopN> be_edcaTxopN = ptr.Get<EdcaTxopN> ();
		 	  be_edcaTxopN->SetMinCw(cwmin);

		 	  rmac->GetAttribute ("BK_EdcaTxopN", ptr);
		 	  Ptr<EdcaTxopN> bk_edcaTxopN = ptr.Get<EdcaTxopN> ();
		 	  bk_edcaTxopN->SetMinCw(cwmin);
		      }
	    }
}

void
Ieee16094Experiment::AssignChannelAccess ()
{
	//		The continuous channel access assigned for default CCH channel
	if (config == CONFIG_CONTINUOUS)
		return;

	if (config == CONFIG_ALTERNATING)
	{
//		The alternating channel access assigned with default 50ms CCHI and 50ms SCHI
//	  Config::SetDefault ("ns3::ChannelCoordinator::CchInterval", TimeValue (MilliSeconds(50)));
//	  Config::SetDefault ("ns3::ChannelCoordinator::SchInterval", TimeValue (MilliSeconds(50)));
	}
	else if (config == CONFIG_ALTERNATING_80_20)
		{
	//		The alternating channel access assigned with 80ms CCHI and 20ms SCHI
		  Config::SetDefault ("ns3::ChannelCoordinator::CchInterval", TimeValue (MilliSeconds(80)));
		  Config::SetDefault ("ns3::ChannelCoordinator::SchInterval", TimeValue (MilliSeconds(20)));
		}
	else if (config == CONFIG_ALTERNATING_20_80)
	{
//		The alternating channel access assigned with 20ms CCHI and 80ms SCHI
	  Config::SetDefault ("ns3::ChannelCoordinator::CchInterval", TimeValue (MilliSeconds(20)));
	  Config::SetDefault ("ns3::ChannelCoordinator::SchInterval", TimeValue (MilliSeconds(80)));
	}
	else
	{
		NS_FATAL_ERROR_NO_MSG();
	}

	  for (uint32_t i = 0; i != devices.GetN (); ++i)
	    {
		  Ptr<WaveNetDevice> device = DynamicCast<WaveNetDevice>(devices.Get(i));

		  // Alternating access without immediate channel switch
		  const SchInfo schInfo = SchInfo (SCH1, false, EXTENDED_ALTERNATING);
		  Simulator::Schedule (Seconds (0.0), &WaveNetDevice::StartSch, device, schInfo);
	    }
}

void
Ieee16094Experiment::SetupMobility ()
{
  // Create a simple highway model based on GridPositionAllocator and ConstantPositionMobilityModel
  MobilityHelper mobility;
  // every node is in 1km * 4line. It is 3 meter between neighboring lines.
  mobility.SetPositionAllocator ("ns3::GridPositionAllocator",
		  	  	  	  	  	  	 "MinX", DoubleValue (0.0),
		  	  	  	  	  	  	 "MinY", DoubleValue (0.0),
		  	  	  	  	  	  	 "DeltaX", DoubleValue (Device_Transmission_Range / (nodesNum / 4.0)),
		  	  	  	  	  	  	 "DeltaY", DoubleValue (3),
		  	  	  	  	  	  	 "GridWidth", UintegerValue (nodesNum / 4),
		  	  	  	  	  	  	 "LayoutType", StringValue ("RowFirst"));
  // here we use constant velocity mobility model for two reasons
  // (a) since the mobility model of ns3 is mainly useful and enough for MANET, we can not ensure
  // whether they are suitable for VANET. Some special mobility models are required for VANET simulation.
  // (b) the mobility characteristic can cause packets lost which we do not want here. Because we want
  // get the impact of the standard on communication performance.
//  mobility.SetMobilityModel ("ns3::ConstantVelocityMobilityModel");
  mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
  mobility.Install (nodes);
  for (NodeContainer::Iterator i = nodes.Begin (); i != nodes.End (); ++i)
    {
	  Ptr<Node> node = (*i);
//	  node->GetObject<ConstantVelocityMobilityModel> ()->SetVelocity (Vector (20, 0, 0));

	  Ptr<MobilityModel> model = node->GetObject<MobilityModel> ();
	  Vector pos = model->GetPosition ();
	  NS_LOG_DEBUG ( "node :" << node->GetId() << " position: " << pos);
    }
}

bool
Ieee16094Experiment::Receive (Ptr<NetDevice> dev, Ptr<const Packet> pkt, uint16_t mode, const Address &sender)
{
  NS_LOG_FUNCTION (this << dev << pkt << mode << sender);
  NS_ASSERT (mode == Packet_Number);

  StatsTag tag;
  bool result;
  result = pkt->FindFirstMatchingByteTag (tag);
  if (!result)
    {
      NS_FATAL_ERROR ("the packet here shall have a stats tag");
    }
  Time now = Now ();
  Time sendTime = tag.GetSendTime ();
  uint32_t packetId = tag.GetPacketId ();
	std::map<uint32_t,  BroadcastPacket *>::iterator i =  broadcastPackets.find(packetId);
	if (i == broadcastPackets.end())
	  {
		NS_FATAL_ERROR_NO_MSG();
		  return true;
	  }
	  BroadcastPacket * packet = i->second;

// get source node id
uint32_t id_src = packet->src;
Ptr<Node> src = NodeList::GetNode(id_src);
Ptr<MobilityModel> model_src = src->GetObject<MobilityModel> ();

// get current node id
Ptr<Node> cur = dev->GetNode();
uint32_t id_cur = cur->GetId();
Ptr<MobilityModel> model_cur = cur->GetObject<MobilityModel> ();

NS_LOG_DEBUG ("Time = " << Now ().GetMicroSeconds() << "us, RECV,   node = " << id_cur << ", packet = " << packetId << " from node = " << id_src << ", at pos = " << model_cur->GetPosition());

// WARNING: there maybe a case (not confirm but possibly) that node A has neighbors B, C and D when broadcasting, however has neighbors B, C and E when receiving neighbors
// The reason maybe the mobility patterns of vehicle nodes.
// The solution here is that we will filter the nodes which will not belong to the neighbor set when broadcasting.
// For example, E will not be used for stats.
if (find (packet->sendNeighbors.begin(), packet->sendNeighbors.end(), id_cur) == packet->sendNeighbors.end ())
{
	  NS_LOG_WARN ("yes this warning case is confirmed");
	  NS_FATAL_ERROR_NO_MSG();
	  return true;
}

NS_ASSERT (packet->sendNeighbors.size() != 0);

packet->recvNeighbors.push_back(id_cur);

// this broadcast packets is received by all neighbors successfully.
if (packet->sendNeighbors.size() == packet->recvNeighbors.size())
{
  	 pdr_sends++;
  	 pdr_recvs++;
  	 pad_delay += (Now() - sendTime).GetMicroSeconds();
  	 prr_sends += packet->sendNeighbors.size();
  	 prr_recvs += packet->recvNeighbors.size();

	// then we should erase this packet record
	  packet->sendNeighbors.clear();
	  packet->recvNeighbors.clear();
	  delete packet;
	  i->second = 0;
	  broadcastPackets.erase(i);
}

  return true;
}

void
Ieee16094Experiment::Send (Ptr<WaveNetDevice> sender, uint32_t size, uint32_t app, uint32_t freq)
{
  NS_LOG_FUNCTION (this << sender);
  Time now = Now ();
  uint32_t packetId = ++sequence;
  Ptr<Packet> packet = Create<Packet> (size);
  StatsTag tag = StatsTag (packetId, now);
  packet->AddByteTag (tag);
  Address dest = Mac48Address::GetBroadcast ();
  TxInfo info = TxInfo (CCH);
  Ptr<Node> src = sender->GetNode();
  uint32_t src_id = src->GetId();
  Ptr<MobilityModel> model_src = src->GetObject<MobilityModel> ();
  Vector pos_src = model_src->GetPosition ();

	  // for stats
	   std::vector<uint32_t> neighbors;
	   for (NodeContainer::Iterator i = nodes.Begin (); i != nodes.End (); ++i)
	  	    {
	  		  if ((*i)->GetId() == src_id)
	  			  continue;
	  		  Ptr<Node> dest = (*i);
	  		  Ptr<MobilityModel> model_dest = dest->GetObject<MobilityModel> ();
	  		  Vector pos_dest = model_dest->GetPosition ();

	  		  if (CalculateDistance (pos_dest, pos_src) <= Device_Transmission_Range )
	  		   {
	  			  neighbors.push_back(dest->GetId());
	  			  continue;
	  		    }
	  	    }
	     if (neighbors.size() == 0)
	     {
	     	NS_LOG_WARN ("Node:" << src_id << " has no neighbors in transmission range");
	     	// this packet will not be used for  stats
	     	return;
	     }

	     BroadcastPacket*  p = new BroadcastPacket();
	     p->src = src_id;
	     p->app = app;
	     p->freq = freq;
	     p->sendNeighbors = neighbors;
	     broadcastPackets.insert(std::make_pair(packetId, p));

  bool result = false;
  result = sender->SendX (packet, dest, app, info);
     if (result)
       NS_LOG_DEBUG ("Time = " << now.GetMicroSeconds () << "us, Send, node =" << src_id << ",  packet = " << tag.GetPacketId ());
     else
       NS_LOG_DEBUG ( "transmit packet fail");

  Ptr<ChannelCoordinator> coordinator = sender->GetChannelCoordinator ();
  if (coordinator->IsCchInterval ())
    {
      if (coordinator->IsGuardInterval ())
        {
          sends.inCgi++;
        }
      else
        {
          sends.inCchi++;
        }
    }
  else
    {
      if (coordinator->IsGuardInterval ())
        {
          sends.inSgi++;
        }
      else
        {
          sends.inSchi++;
        }
    }
}

void
Ieee16094Experiment::StatQueue (Ptr<RegularWifiMac> rmac, std::string queueStr)
{
	  // stat queued packets
      Ptr<WifiMacQueue> queue;
	  PointerValue ptr;
	  rmac->GetAttribute (queueStr, ptr);

	  if (queueStr.compare("DcaTxop") == 0)
	  {
          Ptr<DcaTxop> dcaTxop = ptr.Get<DcaTxop> ();
     	  queue =  dcaTxop->GetQueue();
     	  queues += queue->GetSize();
	  }
	  else
	  {
      	   Ptr<EdcaTxopN> edcaTxopN = ptr.Get<EdcaTxopN> ();
      	   queue =	edcaTxopN->GetEdcaQueue();
      	  queues += queue->GetSize();
	  }

   	  // release queued packets
   	  WifiMacHeader hdr;
   	  Ptr<const Packet> packet;
   	  while((packet = queue->Dequeue(&hdr)))
   	  {
  		   SeqTsHeader seqTs;
   		   ConstCast<Packet>(packet)->RemoveHeader (seqTs);
   		   uint32_t packetId = seqTs.GetSeq ();
   		   broadcastPackets.erase(packetId);
   	  }
}

// when simulation is stopped, we need to stats the queued packets
// so the real transmitted packets will be (sends - queues).
void
Ieee16094Experiment::StatQueuedPackets ()
{
  NS_LOG_FUNCTION (this);
  NetDeviceContainer::Iterator i;
  for (i = devices.Begin (); i != devices.End (); ++i)
    {
	  Ptr<WaveNetDevice> device = DynamicCast<WaveNetDevice>(*i);

      std::map<uint32_t, Ptr<OcbWifiMac> > macs = device->GetMacs ();
        for ( std::map<uint32_t, Ptr<OcbWifiMac> >::iterator k = macs.begin (); k != macs.end (); ++k)
          {
            Ptr<RegularWifiMac> rmac = DynamicCast<RegularWifiMac> (k->second);

            StatQueue (rmac, "DcaTxop");
            StatQueue (rmac, "VO_EdcaTxopN");
            StatQueue (rmac, "VI_EdcaTxopN");
            StatQueue (rmac, "BE_EdcaTxopN");
            StatQueue (rmac, "BK_EdcaTxopN");
          }
      }

  for (std::map<uint32_t, BroadcastPacket *>::iterator j = broadcastPackets.begin (); j!= broadcastPackets.end (); ++j)
  {
	  BroadcastPacket * packet = j->second;
	  NS_ASSERT(packet->sendNeighbors.size() != packet->recvNeighbors.size());

      pdr_sends++;
  	    // this broadcast packet is received partly by its neighbors.
	   prr_sends += packet->sendNeighbors.size();
	   prr_recvs += packet->recvNeighbors.size();

  	  packet->sendNeighbors.clear();
  	  packet->recvNeighbors.clear();
  	  delete packet;
  	  j->second = 0;
    }
	  broadcastPackets.clear();
  }

void
Ieee16094Experiment::InstallApplication (uint32_t app)
{
  NS_LOG_FUNCTION (this);

  // used for sending packets randomly
  Ptr<UniformRandomVariable> rng = CreateObject<UniformRandomVariable> ();
  rng->SetStream (random_stream++);

  double slot = 1.0 / freq;
  for (uint32_t times = 0; times != simulationTime * freq; ++times)
  	  {
		  for (NetDeviceContainer::Iterator i = devices.Begin (); i != devices.End (); ++i)
	  	  {
			  Ptr<WaveNetDevice> sender = DynamicCast<WaveNetDevice> (*i);
			  Time sendTime = Seconds(rng->GetValue(slot * times, slot * (times + 1)));
			  // send event as time 0ms may cause simulation crashed.
			  // I have to avoid this case.
			  while (sendTime.GetMilliSeconds() == 0)
				  sendTime = Seconds(rng->GetValue(slot * times, slot * (times + 1)));
			  Simulator::Schedule (sendTime, &Ieee16094Experiment::Send, this, sender, size, app, freq);
	  	   }
  	  }
}

void
Ieee16094Experiment::Run (void)
{
	 NS_LOG_FUNCTION (this);

	  {
	    pdr = prr = delay = system_throughput = average_throughput = 0.0;
	    for (uint32_t r = 0; r != run; r++)
	    {
	    	RngSeedManager::SetSeed (1);
	    	RngSeedManager::SetRun (17+r);
	    	InitStats ();
	    	CreateNodes ();
	    	SetupMobility ();
	    	CreateWaveDevice ();
	    	AssignChannelAccess ();
	    	InstallApplication (Packet_Number);
	    	Simulator::Stop (Seconds (simulationTime));
	    	Simulator::Run ();
	    	Simulator::Destroy ();
	    	Stats (17+r);
	    }
	    NS_LOG_UNCOND ("PDR   PRR   AverageDelay   SystemThroughput   AverageThroughput");
	    NS_LOG_UNCOND (pdr/run<< " " << prr/run << " " << delay/run << " " << system_throughput/run << " " << average_throughput/run);
	    }
}

void
Ieee16094Experiment::InitStats (void)
{
  std::map<uint32_t, BroadcastPacket *>::iterator i;
  for (i = broadcastPackets.begin(); i != broadcastPackets.end();++i)
  {
	  i->second->sendNeighbors.clear();
	  i->second->recvNeighbors.clear();
	  delete i->second;
	  i->second = 0;
  }
  broadcastPackets.clear ();

  sends = SendStat ();
  sequence = 0;
  queues = 0;
  pdr_sends = 0;
  pdr_recvs = 0;
  pad_delay = 0;
  prr_sends = 0;
  prr_recvs = 0;

  Simulator::ScheduleDestroy (&Ieee16094Experiment::StatQueuedPackets, this);
}

void
Ieee16094Experiment::Stats (uint32_t randomNumber)
{
  NS_LOG_FUNCTION (this);
  if (pdr_recvs == 0)
  {
	  NS_LOG_WARN ("there may be bugs somewhere");
	  return;
  }
  // first show stats records.
  NS_LOG_DEBUG (" simulation result:(run number = " << randomNumber << ")");
  NS_LOG_DEBUG (" nodes = " << nodesNum);
  NS_LOG_DEBUG (" simulation time = " << simulationTime << "s");
  NS_LOG_DEBUG ("      pdr_sends = " << pdr_sends);
  NS_LOG_DEBUG ("      pdr_recvs = " << pdr_recvs);
  NS_LOG_DEBUG ("      pdr_losts = " << (pdr_sends - pdr_recvs)); // the losted packets here contains some partly received packets.
  NS_LOG_DEBUG ("      prr_sends = " << prr_sends);
  NS_LOG_DEBUG ("      prr_recvs = " << prr_recvs);
  NS_LOG_DEBUG ("      prr_losts = " << (prr_sends - prr_recvs));

  // second show performance result
  // stats PDR (packet delivery ratio)
  double PDR = pdr_recvs / (double)(pdr_sends);
  // stats PRR (packet reception ratio)
  double PRR = prr_recvs / (double)(prr_sends);
  // stats average delay
  double AverageDelay = pad_delay / pdr_recvs / 1000.0;
  // stats system throughput, Mbps
  double Throughput = pdr_recvs * size * 8 / simulationTime / 1000.0 / 1000.0;
  // stats average throughput, kbps
  double AverageThroughput = Throughput / (double)nodesNum * 1000;

  NS_LOG_DEBUG (" PDR = " << PDR);
  NS_LOG_DEBUG (" PRR = " << PRR);
  NS_LOG_DEBUG (" AverageDelay = " << AverageDelay << "ms");
  NS_LOG_DEBUG (" Throughput = " << Throughput << "Mbps");
  NS_LOG_DEBUG (" AverageThroughput = " << AverageThroughput << "Kbps");

  pdr+=PDR;
  prr +=PRR;
  delay+=AverageDelay;
  system_throughput+=Throughput;
  average_throughput+=AverageThroughput;
}

int
main (int argc, char *argv[])
{
//  LogComponentEnable ("IEEE16094", LOG_DEBUG);

  Ieee16094Experiment experiment;
  if (experiment.Configure (argc, argv))
    {
      experiment.Run ();
    }
  else
    {
      experiment.Usage ();
    }

  return 0;
}
