/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2005,2006,2007 INRIA, 2015 Konstantinos Katsaros
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 * Mathieu Lacage <mathieu.lacage@sophia.inria.fr>
 * Konstantinos Katsaros <dinos.katsaros@gmail.com>
 */


#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/applications-module.h"
#include "ns3/mobility-module.h"
#include "ns3/wifi-module.h"

#include <iostream>

using namespace ns3;

static void
SetPosition (Ptr<Node> node, Vector position)
{
  Ptr<MobilityModel> mobility = node->GetObject<MobilityModel> ();
  mobility->SetPosition (position);
}

static Vector
GetPosition (Ptr<Node> node)
{
  Ptr<MobilityModel> mobility = node->GetObject<MobilityModel> ();
  return mobility->GetPosition ();
}

static void
AdvancePosition (Ptr<Node> node)
{
  Vector pos = GetPosition (node);
  pos.x += 5.0;
  if (pos.x >= 210.0)
    {
      return;
    }
  SetPosition (node, pos);
  Simulator::Schedule (Seconds (1.0), &AdvancePosition, node);
}

void
TagMarker( Ptr<const Packet> packet)
{
 QosTag qosTag;
 qosTag.SetUserPriority(UP_VO);
 packet->AddPacketTag (qosTag);
}


int main (int argc, char *argv[])
{
  bool g_verbose = true;
  bool infra_mode = false;
  bool qos_flag = false;
  bool adap_rate = false;
  std::string phyMode ("OfdmRate54Mbps");

  CommandLine cmd;
  cmd.AddValue ("verbose", "Print trace information if true", g_verbose);
  cmd.AddValue ("infra", "Infrastructure or Adhoc mode", infra_mode);
  cmd.AddValue ("qos", "Enable QoS-aware MAC", qos_flag);
  cmd.Parse (argc, argv);

  Packet::EnablePrinting ();

  // enable rts cts all the time.
  Config::SetDefault ("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue ("0"));
  // disable fragmentation
  Config::SetDefault ("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue ("2200"));

  NodeContainer stas;
  NodeContainer ap;
  NetDeviceContainer staDevs, apDevs;
  PacketSocketHelper packetSocket;

  stas.Create (2);
  if (infra_mode){
	ap.Create (1);
  }
  // give packet socket powers to nodes.
  packetSocket.Install (stas);
  if (infra_mode){
    packetSocket.Install (ap);
  }

  // mobility.
  MobilityHelper mobility;

  mobility.Install (stas);
  mobility.Install (ap);

  if (infra_mode){
	  Simulator::Schedule (Seconds (1.0), &AdvancePosition, ap.Get (0));
  }

  // Net Devices
  WifiHelper wifi = WifiHelper::Default ();
  wifi.SetStandard (WIFI_PHY_STANDARD_80211a);

  NqosWifiMacHelper NQwifiMac = NqosWifiMacHelper::Default ();
  QosWifiMacHelper QwifiMac = QosWifiMacHelper::Default ();

  YansWifiPhyHelper wifiPhy = YansWifiPhyHelper::Default ();
//  Default Error Model -- NistErrorRateModel

//  For manual WiFi Channel Configuration
//  YansWifiChannelHelper wifiChannel;
//  wifiChannel.SetPropagationDelay ("ns3::ConstantSpeedPropagationDelayModel");
//  wifiChannel.AddPropagationLoss ("ns3::LogDistancePropagationLossModel",
//                                  "Exponent", DoubleValue (3.0));


  YansWifiChannelHelper wifiChannel = YansWifiChannelHelper::Default ();
  wifiPhy.SetChannel (wifiChannel.Create ());

  // ns-3 supports RadioTap and Prism tracing extensions for 802.11
  wifiPhy.SetPcapDataLinkType (YansWifiPhyHelper::DLT_IEEE802_11_RADIO);


  Ssid ssid = Ssid ("wifi-default");

  // Rate Manager
  if (adap_rate){
	  wifi.SetRemoteStationManager ("ns3::ArfWifiManager");
  }
  else{
	  wifi.SetRemoteStationManager ("ns3::ConstantRateWifiManager",
	                                "DataMode",StringValue (phyMode),
	                                "ControlMode",StringValue (phyMode));
  }

  // setup stas.
  if (infra_mode){
	  if (qos_flag){
		 QwifiMac.SetType ("ns3::StaWifiMac",
						   "Ssid", SsidValue (ssid),
						   "ActiveProbing", BooleanValue (false));
		  staDevs = wifi.Install (wifiPhy, QwifiMac, stas);
		  // setup ap.
		  QwifiMac.SetType ("ns3::ApWifiMac",
						   "Ssid", SsidValue (ssid));
		  apDevs = wifi.Install (wifiPhy, QwifiMac, ap);
	  }
	  else{
		 NQwifiMac.SetType ("ns3::StaWifiMac",
						   "Ssid", SsidValue (ssid),
						   "ActiveProbing", BooleanValue (false));
		  staDevs = wifi.Install (wifiPhy, NQwifiMac, stas);
		  // setup ap.
		  NQwifiMac.SetType ("ns3::ApWifiMac",
						   "Ssid", SsidValue (ssid));
		  apDevs = wifi.Install (wifiPhy, NQwifiMac, ap);
	  }
  }
  else {
	  if (qos_flag){
		 QwifiMac.SetType ("ns3::AdhocWifiMac");
		 staDevs = wifi.Install (wifiPhy, QwifiMac, stas);
	  }
	  else{
		 NQwifiMac.SetType ("ns3::AdhocWifiMac");
		 staDevs = wifi.Install (wifiPhy, NQwifiMac, stas);
	  }
  }

  wifiPhy.EnablePcap("train-wifi",staDevs);

  if (infra_mode){
	  wifiPhy.EnablePcap("train-wifi",apDevs);
  }

  PacketSocketAddress psocket;
  psocket.SetSingleDevice (staDevs.Get (0)->GetIfIndex ());
  psocket.SetPhysicalAddress (staDevs.Get (1)->GetAddress ());
  psocket.SetProtocol (1);

  OnOffHelper onoff ("ns3::PacketSocketFactory", Address (psocket));
  onoff.SetConstantRate (DataRate ("500kb/s"));

  ApplicationContainer apps = onoff.Install (stas.Get (0));
  apps.Start (Seconds (0.5));
  apps.Stop (Seconds (43.0));

  if (qos_flag){
	  Config::ConnectWithoutContext ("/NodeList/*/ApplicationList/*/$ns3::OnOffApplication/Tx", MakeCallback (&TagMarker));
  }

  Simulator::Stop (Seconds (44.0));

  Simulator::Run ();

  Simulator::Destroy ();

  return 0;
}
